// [AsmJit]
// Complete JIT Assembler for C++ Language.
//
// [License]
// Zlib - See COPYING file in this package.

// [Guard]
#ifndef _ASMJIT_X86_H
#define _ASMJIT_X86_H

// [Dependencies - AsmJit]
#include "core.h"

#include "x86/x86assembler.h"
#include "x86/x86compiler.h"
#include "x86/x86compilercontext.h"
#include "x86/x86compilerfunc.h"
#include "x86/x86compileritem.h"
#include "x86/x86cpuinfo.h"
#include "x86/x86defs.h"
#include "x86/x86func.h"
#include "x86/x86operand.h"
#include "x86/x86util.h"

// [Guard]
#endif // _ASMJIT_X86_H
