/*
	Copyright (C) 2005-2015 DeSmuME Team
	Copyright 2005-2006 Theo Berkau

	This file is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This file is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the this software.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SNDSDL_H
#define SNDSDL_H

#include "SPU.h"
#define SNDCORE_SDL 2

extern SoundInterface_struct SNDSDL;
#endif
