// [AsmJit]
// Complete JIT Assembler for C++ Language.
//
// [License]
// Zlib - See COPYING file in this package.

#define ASMJIT_EXPORTS

// [Dependencies - AsmJit]
#include "../core/func.h"

// [Api-Begin]
#include "../core/apibegin.h"

namespace AsmJit {
} // AsmJit

// [Api-End]
#include "../core/apiend.h"
